#include <arduino.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <EEPROM.h>
#include "DS_Type.h"
#include "DS_Common.h"
#include "DS_List.h"
#include "DS_DCC.h"
#include "DS_MM2.h"
#include "sseg.h"
#include "DSCore.h"


// Definition constance
#define PROTOCOL ADDR_DCC //Default protocol. If you want to use MM2, set ADDR_MM2.

#define DEBUG

#define BUTTON_DOWN 0b00000001
#define BUTTON_UP   0b00000010
#define BUTTON_F0   0b00000100
#define BUTTON_F1   0b00001000
#define BUTTON_F2   0b00010000
#define BUTTON_DIR  0b00100000

#define WAIT_MSG 50


// Definition for variables
DSCoreLib Gateway;
SSEG SegDisplay(7, 8, A3, A4, 2, 3, 4, A5, A2);

word gLocAddress = 3;
word eepLocAddress = 3;
byte gLocDirection = GO_FWD;
byte gLocSpeed = 0;

byte gButton_val = 0;
byte gLocFunc0 = 0;
byte gLocFunc1 = 0;
byte gLocFunc2 = 0;

byte gDisplayWordCount = 0;
byte gDisplayNext[2] = {SSEG_NONE, SSEG_NONE};

unsigned long gTime_20ms = 0;
unsigned long gTime_100ms = 0;

word gPressCount = 0; /* */


// Functions
byte getButtonStatus();
bool checkButton(byte inBuf, byte inCurrent, byte inButtonBit);
bool checkButtonContinuous(byte inBuf, byte inCurrent, byte inButtonBit);
void saveEEPROM(byte inMode);
void loadEEPROM();
void updateEEPROM();

/*******************************
    Setup
********************************/

void setup()
{
	//Serial
	Serial.begin(115200);
	while (!Serial);

	Serial.println("--------------------------------------");
	Serial.println("Desktop Station Mint Can Box V1.1     ");
	Serial.println("--------------------------------------");
	Serial.println("100 Ready");
	
	//Load EEPROM
	loadEEPROM();
	
	Serial.print("Current Loc Addr: ");
	Serial.println(gLocAddress);
	
	//Analog input
	pinMode(A0, INPUT);
	digitalWrite(A0, LOW); // disable internal pullup
	analogReference(DEFAULT); // set analog reference 5V(VCC) 
	
	//Button
	pinMode(9, INPUT);
	pinMode(10, INPUT);
	pinMode(11, INPUT);
	pinMode(12, INPUT);
	pinMode(13, INPUT);
	pinMode(A1, INPUT);
	digitalWrite(9, HIGH);
	digitalWrite(10, HIGH);
	digitalWrite(11, HIGH);
	digitalWrite(12, HIGH);
	digitalWrite(13, HIGH);
	digitalWrite(A1, HIGH);
	
	//Set default button status
	gButton_val = getButtonStatus();
	
	//Init Gateway
	Gateway.Init();
	
	//Power on
	Gateway.SetPower(true);
	
	//LCD
	SegDisplay.SetWord(SSEG_NONE, SSEG_NONE);
	SegDisplay.SetNumber(gLocAddress);
	SegDisplay.SetInit();

}

/*******************************
    Main loop
********************************/

void loop()
{
	//Scan
	Gateway.Scan();
	
	// 20ms interval
	if( (millis() - gTime_20ms) >= 5)
	{
		//LCD control
		if( gDisplayWordCount > 0)
		{
			SegDisplay.ShowWord();
			gDisplayWordCount--;
			
			//Twice message function
			if( gDisplayWordCount == 0)
			{
				if ((gDisplayNext[0] == SSEG_NONE) && (gDisplayNext[1] == SSEG_NONE))
				{
					// do nothing
				}
				else
				{
					//Display Message
					SegDisplay.SetWord(gDisplayNext[0], gDisplayNext[1]);
					gDisplayWordCount = WAIT_MSG;
					gDisplayNext[0] = SSEG_NONE;
					gDisplayNext[1] = SSEG_NONE;
				}
			}
		}
		else
		{
			SegDisplay.ShowNumber();
		}
		
		//Reset task
		gTime_20ms = millis();
		
	}
	
	// 100ms interval
	if( (millis() - gTime_100ms) >= 100)
	{
		//Dial
		word aSpeed = (analogRead(A0) & 0b1111111100);
		
		if( aSpeed != gLocSpeed)
		{
			gLocSpeed = aSpeed;
			Gateway.SetLocoSpeedEx(PROTOCOL + gLocAddress, aSpeed, 0);
			
			/*
			For MM2 mode
			Gateway.SetLocoSpeedEx(PROTOCOL + gLocAddress, aSpeed, 0);
			*/
			
			//Update EEPROM
			updateEEPROM();
			
			//Debug
			Serial.print("VOL: ");
			Serial.println(aSpeed);
		}
		
		//Buttons
		byte aButton_val = getButtonStatus();
		
		if(aButton_val != gButton_val)
		{
			
			
			if(checkButton(gButton_val, aButton_val, BUTTON_UP))
			{
				//Increment address
				gLocAddress++;
				
				if(gLocAddress >= 100)
				{
					gLocAddress = 1;
				}
				
				//Display Loc No.
				SegDisplay.SetNumber(gLocAddress);
				
				//Reset press count
				gPressCount = 0;
				
				//DEBUG
				Serial.print("ADDR: ");
				Serial.println(gLocAddress);
				
			}
			else if(checkButton(gButton_val, aButton_val, BUTTON_DOWN))
			{
				//Decrement address
				gLocAddress--;
				
				if(gLocAddress <= 0)
				{
					gLocAddress = 99;
				}
				
				//Display Loc No.
				SegDisplay.SetNumber(gLocAddress);
				
				//Reset press count
				gPressCount = 0;
				
				//DEBUG
				Serial.print("ADDR: ");
				Serial.println(gLocAddress);
				
			}
			else if(checkButton(gButton_val, aButton_val, BUTTON_F0))
			{
				
				gLocFunc0 = (gLocFunc0 == 0) ? 1 : 0;
				
				Gateway.SetLocoFunction(PROTOCOL + gLocAddress, 1, gLocFunc0);
				
				//Display Message
				SegDisplay.SetWord(SSEG_F, SSEG_0);
				gDisplayWordCount = WAIT_MSG;
				gDisplayNext[0] = SSEG_O;
				gDisplayNext[1] = (gLocFunc0 == 0) ? SSEG_F : SSEG_N;
				
				//Update EEPROM
				updateEEPROM();

				//DEBUG
				Serial.print("F0: ");
				Serial.println(gLocFunc0);
			}
			else if(checkButton(gButton_val, aButton_val, BUTTON_F1))
			{
				
				gLocFunc1 = (gLocFunc1 == 0) ? 1 : 0;
				
				Gateway.SetLocoFunction(PROTOCOL + gLocAddress, 2, gLocFunc1);
				
				//Display Message
				SegDisplay.SetWord(SSEG_F, SSEG_1);
				gDisplayWordCount = WAIT_MSG;
				gDisplayNext[0] = SSEG_O;
				gDisplayNext[1] = (gLocFunc1 == 0) ? SSEG_F : SSEG_N;
				
				//Update EEPROM
				updateEEPROM();
				
				//DEBUG
				Serial.print("F1: ");
				Serial.println(gLocFunc1);
			}
			else if(checkButton(gButton_val, aButton_val, BUTTON_F2))
			{
				
				gLocFunc2 = (gLocFunc2 == 0) ? 1 : 0;
				
				Gateway.SetLocoFunction(PROTOCOL + gLocAddress, 3, gLocFunc2);
				
				//Update EEPROM
				updateEEPROM();
				
				//Display Message
				SegDisplay.SetWord(SSEG_F, SSEG_2);
				gDisplayWordCount = WAIT_MSG;
				gDisplayNext[0] = SSEG_O;
				gDisplayNext[1] = (gLocFunc2 == 0) ? SSEG_F : SSEG_N;
				
				//DEBUG
				Serial.print("F2: ");
				Serial.println(gLocFunc2);
			}
			else if(checkButton(gButton_val, aButton_val, BUTTON_DIR))
			{
				
				gLocDirection = (gLocDirection == GO_FWD) ? GO_REV : GO_FWD;
				
				Gateway.SetLocoDirection(PROTOCOL + gLocAddress, gLocDirection);
				
				//Display Message
				if( gLocDirection == GO_FWD)
				{
					SegDisplay.SetWord(SSEG_F, SSEG_D);
				}
				else
				{
					SegDisplay.SetWord(SSEG_R, SSEG_V);
				}
				gDisplayWordCount = WAIT_MSG;
				
				//Update EEPROM
				updateEEPROM();
				
				//DEBUG
				Serial.print("DIR: ");
				Serial.println((gLocDirection == GO_FWD) ? "FWD" : "REV");
			}
			
			
			
			//Store to buf
			gButton_val = aButton_val;
			
		}
		else
		{
			//Buttons have no difference from previous state
			
			if(checkButtonContinuous(gButton_val, aButton_val, BUTTON_UP))
			{
				if( gPressCount > 15)
				{
					//Increment address
					gLocAddress++;
					
					if(gLocAddress >= 100)
					{
						gLocAddress = 1;
					}
					
					//Display Loc No.
					SegDisplay.SetNumber(gLocAddress);
				}
				else
				{
					gPressCount++;
				}
			}
			else if(checkButtonContinuous(gButton_val, aButton_val, BUTTON_DOWN))
			{
				if( gPressCount > 15)
				{
					//Decrement address
					gLocAddress--;
					
					if(gLocAddress <= 0)
					{
						gLocAddress = 99;
					}
					
					//Display Loc No.
					SegDisplay.SetNumber(gLocAddress);
				}
				else
				{
					gPressCount++;
				}
			}
		}
		
		//Reset task
		gTime_100ms = millis();
	}
	
}

/*******************************
    getButtonStatus
********************************/

byte getButtonStatus()
{
	
	byte aButton_val = 0;
	
	//Input Buttons
	bitWrite(aButton_val, 0, digitalRead(9)); // Down
	bitWrite(aButton_val, 1, digitalRead(10));// Up
	bitWrite(aButton_val, 2, digitalRead(11));// F0
	bitWrite(aButton_val, 3, digitalRead(12));// F1
	bitWrite(aButton_val, 4, digitalRead(13));// F2
	bitWrite(aButton_val, 5, (digitalRead(A1) == 1) ? 0 : 1);// DIR
	
	return ~aButton_val;
	
}

/*******************************
    checkButton
********************************/

bool checkButton(byte inBuf, byte inCurrent, byte inButtonBit)
{

	if( ((inBuf & inButtonBit) == 0) && ((inCurrent & inButtonBit) > 0))
	{
		return true;
	}
	else
	{
		return false;
	}
	
}

/*******************************
    checkButtonContinuous
********************************/

bool checkButtonContinuous(byte inBuf, byte inCurrent, byte inButtonBit)
{

	if( (inCurrent & inButtonBit) > 0)
	{
		return true;
	}
	else
	{
		return false;
	}
	
}


void updateEEPROM()
{
	if( gLocAddress != eepLocAddress)
	{
		//Update Loc Addr EEPROM
		saveEEPROM(0);
		
		//Update buffer
		eepLocAddress = gLocAddress;
	}
	
}


void loadEEPROM()
{
	
	if( EEPROM.read(0) == 0xCC)
	{
		
		eepLocAddress = (EEPROM.read(3) << 8) + EEPROM.read(2);
		gLocAddress = eepLocAddress;
	}
	
}

void saveEEPROM(byte inMode)
{
	
	if( EEPROM.read(0) == 0xCC)
	{
		
		if( inMode == 0)
		{
			EEPROM.write(2, lowByte(gLocAddress));
			EEPROM.write(3, highByte(gLocAddress));
		}
		
		
		
	}
	else
	{
		//
		EEPROM.write(0, 0xCC);
		
		//AhX
		EEPROM.write(1, 0);
		EEPROM.write(2, lowByte(gLocAddress));
		EEPROM.write(3, highByte(gLocAddress));
		
	}
	
}

